#!/usr/bin/env python
# -*- coding: utf-8 -*-
# template 21
"""
This module extends python os module with a few new functionality
 to interface operating system.
It recognizes file name structure and manages both URI standard name
 both local name, as UNC and ODS5.

- URI (Uniform Resource Identifier) is standard posix filename.
- UNC (Uniform Naming Convention) is windows standard
- ODS5 is used for define OpenVMS standard filenames

An example of URI filename is '/home/myfile'.
UNC example for the same of previous URI name is '\\home\\myfile'
 (with single backslash)
ODS5 (OpenVMS) for the same of previous URI name is '[home]myfile'

See https://en.wikipedia.org/wiki/Path_(computing)
"""
import os
import sys
import pkg_resources
import gzip
import shutil


__version__ = "2.0.7"


def fake_setup(**kwargs):
    globals()["setup_args"] = kwargs


def read_setup():
    setup_info = os.path.abspath(os.path.join(os.path.dirname(__file__), "setup.info"))
    if not os.path.isfile(setup_info):
        setup_info = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "setup.py")
        )
    setup_args = {}
    if os.path.isfile(setup_info):
        with open(setup_info, "r") as fd:
            exec(fd.read().replace("setup(", "fake_setup("))
            setup_args = globals()["setup_args"]
    else:
        print("Not internal configuration file found!")
    setup_args["setup"] = setup_info
    try:
        pkg = pkg_resources.get_distribution(__package__.split(".")[0])
        setup_args["name"] = pkg.key
        setup_args["version"] = pkg.version
    except BaseException:
        pass
    return setup_args


def get_pypi_paths():
    local_venv = "/devel/venv/"
    pkgpath = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
    bin_path = lib_path = ""
    path = pkgpath
    while not bin_path and path != "/" and path != os.environ["HOME"]:
        path = os.path.dirname(path)
        if os.path.isdir(path) and os.path.basename(path) in ("bin", "lib"):
            if (os.path.isdir(os.path.join(os.path.dirname(path), "bin")) and
                    os.path.isdir(os.path.join(os.path.dirname(path), "lib"))):
                bin_path = os.path.join(os.path.dirname(path), "bin")
                lib_path = os.path.join(os.path.dirname(path), "lib")
    if not bin_path and local_venv:
        for path in sys.path:
            if local_venv in path:
                bin_path = os.path.join(
                    path[: path.find(local_venv)],
                    *[x for x in local_venv.split("/") if x][:-1]
                )
                break
    return pkgpath, bin_path, lib_path


def copy_pkg_data(setup_args, verbose):
    if setup_args.get("package_data"):
        pkgpath, bin_path, lib_path = get_pypi_paths()
        if bin_path:
            # TODO> compatibility mode
            bin2_path = os.path.join(os.environ["HOME"], "devel")
            if not os.path.isdir(bin2_path):
                bin2_path = ""
            man_path = os.path.join(bin_path, "man", "man8")
            if not os.path.isdir(man_path):
                man_path = ""
            for pkg in setup_args["package_data"].keys():
                for fn in setup_args["package_data"][pkg]:
                    base = os.path.basename(fn)
                    if base in ("setup.info", "*"):
                        continue
                    full_fn = os.path.abspath(os.path.join(pkgpath, fn))
                    if base.endswith(".man") and man_path:
                        with open(full_fn, "r") as fd:
                            help_text = fd.read()
                        tgt_fn = os.path.join(man_path, "%s.8.gz" % base[:-4])
                        with gzip.open(tgt_fn, "w") as fd:
                            if sys.version_info[0] == 3:
                                fd.write(help_text.encode("utf-8"))
                            else:
                                fd.write(help_text)
                        if verbose:
                            print("$ gzip -c %s > %s" % (full_fn, tgt_fn))
                        continue
                    if lib_path:
                        tgt_fn = os.path.join(lib_path, base)
                        if sys.version_info[0] == 3:
                            try:
                                shutil.copy(full_fn, tgt_fn)
                                if verbose:
                                    print("$ cp %s %s" % (full_fn, tgt_fn))
                            except shutil.SameFileError:
                                pass
                        else:
                            try:
                                shutil.copy(full_fn, tgt_fn)
                                if verbose:
                                    print("$ cp %s %s" % (full_fn, tgt_fn))
                            except BaseException:
                                pass
                    # TODO> compatibility mode
                    tgt_fn = os.path.join(bin_path, base)
                    if os.path.isfile(tgt_fn):
                        os.unlink(tgt_fn)
                    if not os.path.exists(tgt_fn):
                        if verbose:
                            print("$ ln -s %s %s" % (full_fn, tgt_fn))
                        os.symlink(full_fn, tgt_fn)
                    if bin2_path:
                        tgt_fn = os.path.join(bin2_path, base)
                        if os.path.isfile(tgt_fn):
                            os.unlink(tgt_fn)
            # TODO> compatibility mode to remove early
            if lib_path and bin2_path:
                for base in ("z0librc", "odoorc", "travisrc"):
                    full_fn = os.path.join(bin2_path, base)
                    tgt_fn = os.path.join(bin_path, base)
                    if os.path.exists(full_fn) and not os.path.exists(tgt_fn):
                        if verbose:
                            print("$ cp %s %s" % (full_fn, tgt_fn))
                        shutil.copy(full_fn, tgt_fn)


def main(cli_args=None):
    if not cli_args:
        cli_args = sys.argv[1:]
    action = "-H"
    verbose = False
    for arg in cli_args:
        if arg in ("-h", "-H", "--help", "-V", "--version", "--copy-pkg-data"):
            action = arg
        elif arg == "-v":
            verbose = True
    setup_args = read_setup()
    if action == "-h":
        print(
            "%s [-h][-H][--help][-V][--version][-C][--copy-pkg-data]"
            % setup_args["name"]
        )
    elif action in ("-V", "--version"):
        if setup_args["version"] == __version__:
            print(setup_args["version"])
        else:
            print("Version mismatch %s/%s" % (setup_args["version"], __version__))
    elif action in ("-H", "--help"):
        for text in __doc__.split("\n"):
            print(text)
    elif action in ("-C", "--copy-pkg-data"):
        copy_pkg_data(setup_args, verbose)
    return 0



# -*- coding: utf-8 -*-
from . import scripts

from . import clodoo_main as clodoo
from . import transodoo

try:
    from clodoocore import (
        extract_vals_from_rec,
        get_val_from_field,
        cvt_from_ver_2_ver,
        get_model_structure,
        is_valid_field,
        is_required_field,
        model_has_company,
        create_model_object,
    )
except:
    from clodoo.clodoocore import (
        extract_vals_from_rec,
        get_val_from_field,
        cvt_from_ver_2_ver,
        get_model_structure,
        is_valid_field,
        is_required_field,
        model_has_company,
        create_model_object,
    )
try:
    from clodoolib import build_odoo_param, crypt
except:
    from clodoo.clodoolib import build_odoo_param, crypt



# flake8: noqa
# -*- coding: utf-8 -*-
# Import file library
import sys

# import os
import time

try:
    from clodoo import clodoo
except ImportError:
    import clodoo
try:
    from z0lib import z0lib
except ImportError:
    import z0lib
# import pdb


__version__ = "2.0.7"


msg_time = time.time()


def msg_burst(text):
    global msg_time
    t = time.time() - msg_time
    if t > 3:
        print(text)
        msg_time = time.time()


def get_company(ctx):
    model = "res.company"
    ids = clodoo.searchL8(ctx, model, [("name", "ilike", "La Tua Azienda")])
    if not ids:
        ids = clodoo.searchL8(ctx, model, [("id", ">", 1)])
    if ids:
        return ids[0]
    else:
        return 1


def init_n_connect(flavour=None):
    title = "Update Chart of Account"
    parser = z0lib.parseoptargs(
        title, "© 2017-2018 by SHS-AV s.r.l.", version=__version__
    )
    parser.add_argument("-h")
    parser.add_argument(
        "-c",
        "--config",
        help="configuration command file",
        dest="conf_fn",
        metavar="file",
        default="./import_partners.conf",
    )
    parser.add_argument(
        "-d", "--dbname", help="DB name", dest="db_name", metavar="file", default=""
    )
    parser.add_argument("-n")
    parser.add_argument("-q")
    parser.add_argument("-V")
    parser.add_argument("-v")
    # Connect to DB
    ctx = parser.parseoptargs(sys.argv[1:], apply_conf=False)
    oerp, uid, ctx = clodoo.oerp_set_env(
        confn=ctx["conf_fn"], db=ctx["db_name"], ctx=ctx
    )
    return oerp, uid, ctx


def get_company_account(ctx):
    company_id = ctx["company_id"]
    ids = clodoo.searchL8(
        ctx, "account.account", [("code", "=", "0"), ("company_id", "=", company_id)]
    )
    if len(ids) != 1:
        print("Company account not found!")
        sys.exit(1)
    return ids[0]


def get_code_values(ctx, code):
    model = "account.account"
    company_id = ctx["company_id"]
    vals = {}
    vals["code"] = code
    root_id = get_company_account(ctx)
    vals["parent_id"] = root_id
    vals["company_id"] = company_id
    vals["type"] = "view"
    ids = clodoo.searchL8(ctx, "account.account.type", [("code", "=", "view")])
    if len(ids) == 0:
        print("Type view not found!")
        sys.exit(1)
    vals["user_type"] = ids[0]
    if code == "A":
        vals["name"] = "ATTIVITÀ"
    elif code == "P":
        vals["name"] = "PASSIVITÀ"
    elif code == "R":
        vals["name"] = "RICAVI"
        ids = clodoo.searchL8(
            ctx, model, [("code", "=", "__"), ("company_id", "=", company_id)]
        )
        if len(ids) != 1:
            print("L&P account not found!")
            sys.exit(1)
        vals["parent_id"] = ids[0]
    elif code == "S":
        vals["name"] = "COSTI"
        ids = clodoo.searchL8(
            ctx, model, [("code", "=", "__"), ("company_id", "=", company_id)]
        )
        if len(ids) != 1:
            print("L&P account not found!")
            sys.exit(1)
        vals["parent_id"] = ids[0]
    elif code == "__":
        vals["name"] = "UTILE (-) O PERDITA (+) DI ESERCIZIO"
    else:
        print("Invalid account code %s" % code)
        sys.exit(1)
    return vals


def update_coa(ctx):
    company_id = ctx["company_id"]
    model = "account.account"
    if len(clodoo.searchL8(ctx, model, [("company_id", "=", company_id)])) < 100:
        return
    company = clodoo.browseL8(ctx, "res.company", company_id)
    print("- Processing company %s" % company.name)
    for code in ("__", "A", "P", "R", "S"):
        ids = clodoo.searchL8(
            ctx, model, [("code", "=", code), ("company_id", "=", company_id)]
        )
        if len(ids) > 2:
            print("Warning: invalid account '%s'" % code)
            sys.exit(1)
        elif ids:
            vals = get_code_values(ctx, code)
            try:
                clodoo.writeL8(ctx, model, ids[0], vals)
            except BaseException:
                pass
        else:
            vals = get_code_values(ctx, code)
            clodoo.createL8(ctx, model, vals)
    aprs = {}
    for code in ("__", "A", "P", "R", "S"):
        ids = clodoo.searchL8(
            ctx, model, [("code", "=", code), ("company_id", "=", company_id)]
        )
        aprs[code] = ids[0]
    root_id = get_company_account(ctx)
    account_ids = clodoo.searchL8(
        ctx,
        model,
        [
            ("type", "=", "view"),
            ("parent_id", "=", root_id),
            ("id", "not in", aprs.values()),
        ],
    )
    for account_id in account_ids:
        ids = clodoo.searchL8(ctx, model, [("parent_id", "=", account_id)])
        if ids:
            account = clodoo.browseL8(ctx, model, ids[0])
            while account.type == "view":
                ids = clodoo.searchL8(ctx, model, [("parent_id", "=", ids[0])])
                if ids:
                    account = clodoo.browseL8(ctx, model, ids[0])
                else:
                    account.type = "asset"
            if account.user_type.code in ("asset", "bank", "cash", "receivable"):
                parent_id = aprs["A"]
            elif account.user_type.code in ("equity", "liability", "payable", "tax"):
                parent_id = aprs["P"]
            elif account.user_type.code == "income":
                parent_id = aprs["R"]
            elif account.user_type.code == "expense":
                parent_id = aprs["S"]
            else:
                parent_id = False
            if parent_id:
                vals = {}
                vals["parent_id"] = parent_id
                clodoo.writeL8(ctx, model, account_id, vals)


if __name__ == "__main__":
    # pdb.set_trace()
    oerp, uid, ctx = init_n_connect()
    print("Update Chart of Account on DB %s" % (ctx["db_name"]))
    ids = clodoo.searchL8(ctx, "res.company", [])
    for company_id in ids:
        ctx["company_id"] = company_id
        update_coa(ctx)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright SHS-AV s.r.l. <http://ww.zeroincombenze.it>)
#
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).
#
#    All Rights Reserved
#
"""Translate Odoo name to migrate from version to another
Structure:
[pymodel]:       Odoo model name
[pymodel][ttype]: may be:
    name:     symbolic name of a field (deprecated)
    field:    field name to translate
    action:   action/function to translate
    xref:     external reference (model is ir.model.data)
    model:    model name to translate (model is ir.model)
    module:   module name to translate (model is ir.module.module)
    merge:    module name or model merged with
    value:    value of field to translate (name is the field name)
    valuetnl: field ha translation ("1")

[pymodel][ttype][hash]           hash entry with ver.name list
[pymodel][ttype][hash][ver.name] value for specific version
[pymodel][ttype][ver.name]       specific name entry for name.ver -> hash

the ttype 'value' has a more level for every field name:
[pymodel]['value'][fldname][hash]
[pymodel]['value'][fldname][hash][ver.name]
[pymodel]['value'][fldname][ver.name]
"""
from __future__ import print_function, unicode_literals

import os
import re
import sys

from openpyxl import Workbook, load_workbook
from openpyxl.styles import Font, PatternFill
from openpyxl.utils import get_column_letter
from past.builtins import basestring

from python_plus import bstrings, unicodes

try:
    import jsonlib
except ImportError:
    pass
try:
    from z0lib.z0lib import z0lib
except ImportError:
    try:
        from z0lib import z0lib
    except ImportError:
        import z0lib

__version__ = "2.0.7"
VERSIONS = [
    "6.1",
    "7.0",
    "8.0",
    "9.0",
    "10.0",
    "11.0",
    "12.0",
    "13.0",
    "14.0",
    "15.0",
    "16.0",
]
ALL_VERSIONS = [x for x in VERSIONS]
for org in ("zero", "powerp", "librerp"):
    for ver in VERSIONS:
        if org == "librerp" and ver not in ("6.1", "12.0"):
            continue
        elif org == "powerp" and int(ver.split(".")[0]) < 12:
            continue
        ALL_VERSIONS.append("%s%s" % (org, ver.split(".")[0]))
CVT_ACC_TYPE_OLD_NEW = {
    "Bank": "Bank and Cash",
    "Cash": "Bank and Cash",
    "Check": "Credit Card",
    "Asset": "Current Assets",
    "Liability": "Current Liabilities",
    "Tax": "Current Liabilities",
}
CVT_ACC_TYPE_NEW_OLD = {
    "Bank and Cash": "Bank",
    "Credit Card": "Check",
    "Current Assets": "Asset",
    "Non-current Assets": "Asset",
    "Fixed Asset": "Asset",
    "Current Liabilities": "Liability",
    "Non-current Liabilities": "Liability",
    "Other Income": "Income",
    "Depreciation": "Expense",
    "Cost of Revenue": "Expense",
    "Prepayments": "Expense",
    "Current Year Earnings": "Expense",
}


def natstr(s):
    if sys.version_info[0] == 2:
        return bstrings(s)
    return unicodes(s)


def get_pymodel(model, ttype=None):
    return (
        {"xref": "ir.model.data", "model": "ir.model", "module": "ir.module.module"}
        .get(ttype, model or "res.groups")
        .lower()
    )


def get_ver_name(name, cur_ver):
    if name:
        return "%s~%s" % (cur_ver, name)
    return name


def is_hash(name):
    if re.match(r"^[a-zA-Z0-9_]+[a-zA-Z0-9_.-]*[0-9]+~", name):
        return False
    return True


def set_hash(ttype, name, ver_names):
    if ttype == "name":
        return name.upper()
    key = name if ttype in ("value", "valuetnl") else ""
    for item in ver_names:
        if key:
            key = "%s|%s" % (key, item)
        else:
            key = item
    if ttype in ("value", "valuetnl"):
        return key
    return key.upper()


def build_alias_struct(mindroot, model, ttype, fld_name=False):
    pymodel = get_pymodel(model, ttype=ttype)
    mindroot[pymodel] = mindroot.get(pymodel, {})
    mindroot[pymodel][ttype] = mindroot[pymodel].get(ttype, {})
    if ttype in ("value", "valuetnl") and fld_name:
        mindroot[pymodel][ttype][fld_name] = mindroot[pymodel][ttype].get(fld_name, {})
    return mindroot


def link_versioned_name(mindroot, model, hashed, ttype, src_name, ver, fld_name=False):
    pymodel = get_pymodel(model, ttype=ttype)
    ver_name = get_ver_name(src_name, ver)
    if ttype in ("value", "valuetnl"):
        if src_name and src_name.startswith("${") and src_name.endswith("}"):
            mindroot[pymodel][ttype][fld_name] = src_name
            item = None
        else:
            item = mindroot[pymodel][ttype][fld_name]
    else:
        item = mindroot[pymodel][ttype]
    if item is not None:
        item[hashed] = item.get(hashed, {})
        if ver_name:
            tver = ver_name
            while tver:
                tver = previous_ver_name(tver, ver_name)
                if tver in item and hashed not in item[tver]:
                    item[ver_name] = item[tver]
            if ver_name in item:
                if not isinstance(item[ver_name], list):
                    item[ver_name] = [item[ver_name]]
                if hashed not in item[ver_name]:
                    item[ver_name].append(hashed)
            else:
                item[ver_name] = hashed
        item[hashed][ver] = src_name
    return mindroot


def get_majver(name):
    x = re.search(r"[0-9]+", name)
    if x:
        return int(name[x.start() : x.end()])
    return 0


def tnl_by_code(ctx, model, src_name, src_ver, tgt_ver, name):
    src_majver = get_majver(src_ver)
    tgt_majver = get_majver(tgt_ver)
    if name == "${amount}":
        if isinstance(src_name, basestring):
            src_name = float(src_name)
        if (
            src_majver < 9
            and tgt_majver >= 9
            and src_name in (0.04, 0.05, 0.1, 0.21, 0.22)
        ):
            name = src_name * 100
        elif src_majver >= 9 and tgt_majver < 9 and src_name in (4, 5, 10, 21, 22):
            name = src_name / 100
        else:
            name = src_name
    elif name == "${coa}":
        if tgt_ver.split(".")[0].isdigit() and (
            src_ver.startswith("librerp")
            or src_ver.startswith("powerp")
            or src_ver.startswith("zero")
        ):
            name = {
                "121100": "1101",
                "121330": "1106",
                "121500": "1108",
                "122100": "1111",
                "122330": "1116",
                "123340": "1207",
                "122500": "1118",
                "123380": "1205",
                "123390": "1206",
                "124100": "1211",
                "124200": "1212",
                "124300": "1214",
                "124340": "1217",
                "124380": "1215",
                "124390": "1216",
                "152100": "1501",
                "152150": "1509",
                "152200": "1505",
                "152220": "1506",
                "153010": "1601",
                "153020": "1605",
                "153030": "1613",
                "153050": "1611",
                "153060": "1612",
                "153070": "1602",
                "153110": "1609",
                "190110": "1901",
                "190120": "1902",
                "211010": "2101",
                "219010": "2102",
                "240100": "2420",
                "246000": "2410",
                "250100": "2501",
                "260010": "2601",
                "260050": "2612",
                "260060": "2611",
                "260110": "2602",
                "290010": "2701",
                "290020": "2702",
                "490010": "2801",
                "490020": "2802",
                "490050": "2950",
                "510000": "3112",
                "510100": "3101",
                "510200": "3202",
                "512000": "3103",
                "550530": "3210",
                "610100": "4101",
                "610110": "4102",
                "621120": "4213",
                "621200": "4105",
                "621340": "4201",
                "623100": "4209",
                "623460": "4204",
                "623580": "4203",
                "625010": "4216",
                "625020": "4206",
                "625060": "4205",
                "625100": "4202",
                "630100": "4301",
                "640910": "4401",
                "731140": "5203",
            }.get(src_name, src_name)
        elif src_ver.split(".")[0].isdigit() and (
            tgt_ver.startswith("librerp")
            or tgt_ver.startswith("powerp")
            or tgt_ver.startswith("zero")
        ):
            name = {
                "1101": "121100",
                "1106": "121330",
                "1108": "121500",
                "1111": "122100",
                "1116": "122330",
                "1118": "122500",
                "1205": "123380",
                "1206": "123390",
                "1207": "123340",
                "1211": "124100",
                "1212": "124200",
                "1214": "124300",
                "1215": "124380",
                "1216": "124390",
                "1217": "124340",
                "1501": "152100",
                "1505": "152200",
                "1506": "152220",
                "1509": "152150",
                "1601": "153010",
                "1602": "153070",
                "1605": "153020",
                "1609": "153110",
                "1611": "153050",
                "1612": "153060",
                "1613": "153030",
                "1901": "190110",
                "1902": "190120",
                "2101": "211010",
                "2102": "219010",
                "2410": "246000",
                "2420": "240100",
                "2501": "250100",
                "2601": "260010",
                "2602": "260110",
                "2611": "260050",
                "2612": "260060",
                "2701": "290010",
                "2702": "290020",
                "2801": "490010",
                "2802": "490020",
                "2950": "490050",
                "3101": "510100",
                "3103": "512000",
                "3112": "510000",
                "3202": "510200",
                "3210": "550530",
                "4101": "610100",
                "4102": "610110",
                "4105": "621200",
                "4201": "621340",
                "4202": "625100",
                "4203": "623580",
                "4204": "623460",
                "4205": "625060",
                "4206": "625020",
                "4209": "623100",
                "4213": "621120",
                "4216": "625010",
                "4301": "630100",
                "4401": "640910",
                "5203": "731140",
            }.get(src_name, src_name)
        else:
            name = src_name
    elif name == "${tax}":
        if tgt_ver.split(".")[0].isdigit() and (
            src_ver.startswith("librerp")
            or src_ver.startswith("powerp")
            or src_ver.startswith("zero")
        ):
            name = {"a15a": "00art15a", "a15v": "00art15v"}.get(src_name, src_name)
        elif src_ver.split(".")[0].isdigit() and (
            tgt_ver.startswith("librerp")
            or tgt_ver.startswith("powerp")
            or tgt_ver.startswith("zero")
        ):
            name = {"00art15a": "a15a", "00art15v": "a15v"}.get(src_name, src_name)
        else:
            name = src_name
    return name


def previous_ver_name(ver_name, orig_name):
    x = re.search("[0-9]+", ver_name)
    if x:
        version = int(ver_name[x.start() : x.end()]) - 1
        if version < 6:
            if re.match(r"[0-9]+\.[0-9]", ver_name):
                ver_name = ""
            else:
                ver_name = orig_name if orig_name else ""
            x = re.search("[0-9]+", ver_name)
            if x:
                version = int(ver_name[x.start() : x.end()])
                ver_name = "%d.0" % version
                if "~" in orig_name:
                    ver_name = "%s~%s" % (ver_name, orig_name.split("~", 1)[-1])
            else:
                ver_name = ""
        else:
            ver_name = "%s%d%s" % (
                ver_name[0 : x.start()],
                version,
                ver_name[x.end() :],
            )
        if "6.0" in ver_name:
            ver_name = ver_name.replace("6.0", "6.1")
    return ver_name


def translate_from_to(
    ctx, model, src_name, src_ver, tgt_ver, ttype=False, fld_name=False, type=None
):
    """Translate symbol <src_name> from <src_ver> to <tgt_ver> of Odoo.
    If ttype not supplied, translation is applied for 'name' and 'field' ttypes
    If ttype is in ('value', 'valuetnl'), the param <fld_name> must by supplied.
    Param type is deprecated. It used just for compatibility with old version
    """
    if not ttype and type:
        ttype = type
    del type
    mindroot = ctx.get("mindroot", {})
    if src_ver not in ALL_VERSIONS:
        print("Invalid source version!")
        return ""
    if tgt_ver not in ALL_VERSIONS:
        print("Invalid target version!")
        return ""
    if ttype in ("value", "valuetnl") and not fld_name:
        print("Translation of value require field name!")
        return ""
    if ttype == "valuetnl" and not src_name:
        src_name = "dummy"
    pymodel = get_pymodel(model, ttype=ttype)
    ver_name = get_ver_name(src_name, src_ver)
    name = src_name
    if ver_name and pymodel in mindroot:
        names = []
        for typ in (
            map(lambda x: x, ("name", "field")) if not ttype else [ttype, "merge"]
        ):
            if ttype == "valuetnl":
                if fld_name in mindroot[pymodel].get("value", {}):
                    names = ["1"]
                else:
                    names = [""]
                break
            elif ttype == "value":
                item = mindroot[pymodel].get(typ, {}).get(fld_name, {})
            else:
                item = mindroot[pymodel].get(typ, {})
            if isinstance(item, basestring):
                if item.startswith("${") and item.endswith("}"):
                    # fct = item[2: -1]
                    names.append(
                        tnl_by_code(ctx, model, src_name, src_ver, tgt_ver, item)
                    )
            else:
                sver = ver_name
                while sver and sver not in item:
                    sver = previous_ver_name(sver, ver_name)
                if sver:
                    hashed = item[sver]
                    if not isinstance(hashed, list):
                        hashed = [hashed]
                    for hh in hashed:
                        if hh in item:
                            tver = tgt_ver
                            while tver and tver not in item[hh]:
                                tver = previous_ver_name(tver, tgt_ver)
                            if tver and item[hh][tver] not in names:
                                names.append(item[hh][tver])
                            if (
                                typ == "merge"
                                and src_name not in names
                                and get_majver(tver) < get_majver(sver)
                            ):
                                names.insert(0, src_name)
        if names:
            if len(names) == 1:
                name = names[0]
            else:
                name = names
        if name == r"\N":
            name = None
    return name


def translate_from_sym(ctx, model, sym, tgt_ver):
    # deprecated!
    mindroot = ctx.get("mindroot", {})
    if tgt_ver not in ALL_VERSIONS:
        print("Invalid target version!")
        return ""
    pymodel = get_pymodel(model)
    name = ""
    for typ in ("name", "field"):
        if (
            pymodel in mindroot
            and typ in mindroot[pymodel]
            and sym in mindroot[pymodel][typ]
        ):
            tver = tgt_ver
            while tver and tver not in mindroot[pymodel][typ][sym]:
                tver = previous_ver_name(tver, tgt_ver)
            if tver:
                name = mindroot[pymodel][typ][sym][tver]
                break
    return name


def model_info(ctx, pymodel, tgt_ver, ttype=False, fld_name=False):
    def do_line_ver(mindroot, model, name, typ, fld_name, tgt_ver):
        if typ not in info:
            info[model][typ] = {}
        for ver in ALL_VERSIONS:
            if tgt_ver and ver != tgt_ver:
                continue
            if ver not in info:
                info[model][typ][ver] = []
            if typ == "value":
                if ver in mindroot[model][typ][fld_name][name]:
                    info[model][typ][ver].append(
                        mindroot[model][typ][fld_name][name][ver]
                    )
            else:
                tver = ver
                while tver and tver not in mindroot[model][typ][name]:
                    tver = previous_ver_name(tver, ver)
                if tver in mindroot[model][typ][name]:
                    info[model][typ][ver].append(mindroot[model][typ][name][tver])
        return info

    mindroot = ctx.get("mindroot", {})
    if tgt_ver and tgt_ver not in ALL_VERSIONS:
        print("Invalid source version!")
        return ""
    if ttype in ("value", "valuetnl") and not fld_name:
        print("Translation of value require field name!")
        return ""
    info = {}
    for model in mindroot.keys():
        if pymodel and model != pymodel:
            continue
        if model not in info:
            info[model] = {}
        for typ in mindroot[model]:
            if ttype and typ != ttype:
                continue
            if typ == "value":
                for fld_name in mindroot[model][typ]:
                    for name in mindroot[model][typ][fld_name]:
                        if not is_hash(name):
                            continue
                        do_line_ver(mindroot, model, name, typ, fld_name, tgt_ver)
            else:
                for name in mindroot[model][typ]:
                    if not is_hash(name):
                        continue
                    do_line_ver(mindroot, model, name, typ, False, tgt_ver)
    if pymodel:
        info = info.get(pymodel, {})
        if ttype:
            info = info.get(ttype, {})
            if tgt_ver:
                info = info.get(tgt_ver, [])
    return info


def read_stored_dict(ctx):
    if "mindroot" in ctx:
        return
    if "dict_fn" not in ctx or not ctx["dict_fn"]:
        p = os.path.dirname(__file__) or "."
        if os.path.isfile("%s/transodoo.xlsx" % p):
            ctx["dict_fn"] = "%s/transodoo.xlsx" % p
        elif os.path.isfile(os.path.join(os.path.expanduser("~"), "transodoo.xlsx")):
            ctx["dict_fn"] = os.path.join(os.path.expanduser("~"), "transodoo.xlsx")
        else:
            ctx["dict_fn"] = "transodoo.xlsx"
    mindroot = {}
    wb = load_workbook(ctx["dict_fn"])
    sheet = wb.active
    colnames = []
    for column in sheet.columns:
        colnames.append(column[0].value)
    hdr = True
    for line in sheet.rows:
        if hdr:
            hdr = False
            continue
        row = {}
        for column, cell in enumerate(line):
            row[colnames[column]] = cell.value
        mindroot = build_alias_struct(
            mindroot, row["model"], row["type"], fld_name=row["name"]
        )
        ver_names = []
        used_versions = []
        last_ver = ""
        last_ver_value = ""
        for ver in ALL_VERSIONS:
            if (
                (ver.startswith("zero") and not last_ver.startswith("zero"))
                or (ver.startswith("powerp") and not last_ver.startswith("powerp"))
                or (ver.startswith("librerp") and not last_ver.startswith("librerp"))
            ):
                last_ver_value = ""
            if ver in row and row[ver] and row[ver] != last_ver_value:
                ver_names.append(row[ver])
                used_versions.append(ver)
                last_ver_value = row[ver]
            last_ver = ver
        hashed = set_hash(row["type"], row["name"], ver_names)
        for ver in used_versions:
            mindroot = link_versioned_name(
                mindroot,
                row["model"],
                hashed,
                row["type"],
                row[ver],
                ver,
                fld_name=row["name"],
            )
    ctx["mindroot"] = mindroot


def write_stored_dict(ctx):
    def write_row(sheet, line, rowid, header, widths):
        rowid += 1
        fill = "FFFFFF" if rowid % 2 else "F0F0F0"
        if not line:
            for col, item in enumerate(header):
                font_color = "000000" if col % 2 else "1F1F1F"
                cell = sheet.cell(row=rowid, column=col + 1)
                cell.value = item
                cell.font = Font(name="arial", size=8, color=font_color)
                # cell.fill = PatternFill(bgColor=fill, fill_type="solid")
                cell.fill = PatternFill(bgColor=fill)
                widths[item] = len(item)
        else:
            for col, item in enumerate(header):
                font_color = "000000" if col % 2 else "1F1F1F"
                if item in line:
                    cell = sheet.cell(row=rowid, column=col + 1)
                    cell.value = line[item]
                    cell.font = Font(name="arial", size=8, color=font_color)
                    # cell.fill = PatternFill(bgColor=fill, fill_type="solid")
                    cell.fill = PatternFill(bgColor=fill)
                    widths[item] = min(24, max(len(line[item]), widths[item]))
        return rowid

    # ctx['dict_fn'] = os.path.join(os.path.expanduser('~/transodoo.xlsx'))
    wb = Workbook()
    sheet = wb.active
    # for sheet in wb:
    #     break
    sheet.font = Font(name="arial", size=8)
    sheet.title = "transodoo"
    mindroot = ctx["mindroot"]
    rowid = 0
    header = ["model", "name", "type"]
    for ver in ALL_VERSIONS:
        header.append(ver)
    widths = {}
    rowid = write_row(sheet, None, rowid, header, widths)
    for model in sorted(mindroot.keys()):
        for ttype in sorted(mindroot[model].keys()):
            if ttype in ("value", "valuetnl"):
                iterate = sorted(mindroot[model][ttype].keys())
            else:
                iterate = [None]
            for name in iterate:
                if ttype in ("value", "valuetnl"):
                    items = mindroot[model][ttype][name]
                else:
                    items = mindroot[model][ttype]
                if isinstance(items, basestring):
                    iterate2 = [items]
                else:
                    iterate2 = sorted(items.keys())
                for hashed in iterate2:
                    if not is_hash(hashed):
                        continue
                    line = {"model": model, "type": ttype}
                    if ttype == "name":
                        line["name"] = hashed
                    else:
                        if ttype in ("value", "valuetnl"):
                            line["name"] = name
                    if isinstance(items, basestring):
                        for ver_name in VERSIONS:
                            line[ver_name] = hashed
                    else:
                        for ver_name in sorted(items[hashed].keys()):
                            if items[hashed][ver_name]:
                                line[ver_name] = items[hashed][ver_name]
                    rowid = write_row(sheet, line, rowid, header, widths)
    for col, item in enumerate(header):
        sheet.column_dimensions[get_column_letter(col + 1)].width = widths[item]
    bup_fn = "%s.bak" % ctx["dict_fn"]
    if os.path.isfile(bup_fn):
        os.unlink(bup_fn)
    os.rename(ctx["dict_fn"], bup_fn)
    wb.save(filename=ctx["dict_fn"])


def cvt_file(ctx):
    def write_row(sheet, line, rowid, header, widths):
        rowid += 1
        if not line:
            for col, item in enumerate(header):
                cell = sheet.cell(row=rowid, column=col + 1)
                cell.value = item
                cell.font = Font(name="arial", size=9)
                widths[item] = len(item)
        else:
            for col, item in enumerate(header):
                if item in line:
                    cell = sheet.cell(row=rowid, column=col + 1)
                    cell.value = line[item]
                    cell.font = Font(name="arial", size=9)
                    if line[item] and isinstance(line[item], basestring):
                        widths[item] = min(24, max(len(line[item]), widths[item]))
        return rowid

    if not ctx["file_2_cvt"]:
        print("Missed filename!\nuse -F switch")
        return 1
    wbi = load_workbook(ctx["file_2_cvt"])
    sheeti = wbi.active
    # for sheeti in wbi:
    #     break
    wbo = Workbook()
    sheeto = wbo.active
    # for sheeto in wbo:
    #     break
    sheeto.font = Font(name="arial", size=9)
    colnames = []
    for column in sheeti.columns:
        colnames.append(column[0].value)
    hdr = True
    rowid = 0
    widths = {}
    for line in sheeti.rows:
        if hdr:
            hdr = False
            new_colnames = []
            tnl_colname = {}
            for column in colnames:
                tnl_colname[column] = translate_from_to(
                    ctx,
                    ctx["model"],
                    column,
                    ctx["oe_from_ver"],
                    ctx["odoo_ver"],
                    ttype="name",
                )
                new_colnames.append(tnl_colname[column])
            rowid = write_row(sheeto, None, rowid, new_colnames, widths)
            continue
        row = {}
        for column, cell in enumerate(line):
            row[colnames[column]] = cell.value
        new_row = {}
        for item in row.keys():
            new_row[tnl_colname[item]] = translate_from_to(
                ctx,
                ctx["model"],
                row[item],
                ctx["oe_from_ver"],
                ctx["odoo_ver"],
                ttype="value",
                fld_name=item,
            )
        rowid = write_row(sheeto, new_row, rowid, new_colnames, widths)
    for col, item in enumerate(new_colnames):
        sheeto.column_dimensions[get_column_letter(col + 1)].width = widths[item]
    bup_fn = "%s.bak" % ctx["file_2_cvt"]
    if os.path.isfile(bup_fn):
        os.unlink(bup_fn)
    os.rename(ctx["file_2_cvt"], bup_fn)
    wbo.save(filename=ctx["file_2_cvt"])


def transodoo_list(ctx):
    if ctx["model"]:
        pymodel = get_pymodel(ctx["model"], ttype=ctx["opt_kind"])
    else:
        pymodel = ""
    info = model_info(
        ctx, pymodel, ctx["odoo_ver"], ttype=ctx["opt_kind"], fld_name=ctx["field_name"]
    )
    if ctx["model"]:
        print("Model: %s" % get_pymodel(ctx["model"]))
        if ctx["opt_kind"]:
            print("- Type: %s" % ctx["opt_kind"])
            if ctx["odoo_ver"]:
                print("  - Odoo version: %s" % ctx["odoo_ver"])
    if info:
        print(jsonlib.write(info, indent="    "))


def transodoo(ctx=None):
    if ctx["action"] == "list":
        read_stored_dict(ctx)
        transodoo_list(ctx)
    elif ctx["action"] == "translate":
        read_stored_dict(ctx)
        if ctx["oe_from_ver"]:
            print(
                natstr(
                    translate_from_to(
                        ctx,
                        ctx["model"],
                        ctx["sym"],
                        ctx["oe_from_ver"],
                        ctx["odoo_ver"],
                        ttype=ctx["opt_kind"],
                        fld_name=ctx["field_name"],
                    )
                )
            )
        else:
            print(
                natstr(
                    translate_from_sym(ctx, ctx["model"], ctx["sym"], ctx["odoo_ver"])
                )
            )
    elif ctx["action"] == "rewrite":
        read_stored_dict(ctx)
        write_stored_dict(ctx)
    elif ctx["action"] == "cvt-file":
        read_stored_dict(ctx)
        cvt_file(ctx)
    else:
        print("Invalid action!")
        return 1
    return 0


def main(cli_args=None):
    if not cli_args:
        cli_args = sys.argv[1:]
    parser = z0lib.parseoptargs(
        "Transodoo", "© 2017-2019 by SHS-AV s.r.l.", version=__version__
    )
    parser.add_argument("-h")
    parser.add_argument("-b", "--branch", action="store", dest="odoo_ver", default="")
    parser.add_argument(
        "-F", "--filename", action="store", dest="file_2_cvt", default=""
    )
    parser.add_argument(
        "-f", "--from-branch", action="store", dest="oe_from_ver", default=""
    )
    parser.add_argument(
        "-k", "--kind", action="store", dest="opt_kind", default="field"
    )
    parser.add_argument(
        "-l", "--language", action="store", dest="opt_lang", default="it_IT"
    )
    parser.add_argument("-m", "--model", action="store", dest="model")
    parser.add_argument("-N", "--field-name", action="store", dest="field_name")
    parser.add_argument("-n")
    parser.add_argument("-q")
    parser.add_argument("-s", "--symbol-value", action="store", dest="sym", default="")
    parser.add_argument("-V")
    parser.add_argument("-v")
    parser.add_argument("action", help="list,translate,rewrite")
    ctx = parser.parseoptargs(sys.argv[1:])
    if ctx["odoo_ver"]:
        if ctx["odoo_ver"] not in ALL_VERSIONS:
            print(
                "Invalid version %s!\nUse one of %s" % (ctx["odoo_ver"], ALL_VERSIONS)
            )
            sys.exit(1)
    if ctx["oe_from_ver"]:
        if ctx["oe_from_ver"] not in ALL_VERSIONS:
            print(
                "Invalid version %s!\nUse one of %s" % (ctx["odoo_ver"], ALL_VERSIONS)
            )
            sys.exit(1)
    sts = transodoo(ctx=ctx)
    return sts


if __name__ == "__main__":
    exit(main())



# -*- coding: utf-8 -*-
#
# Copyright SHS-AV s.r.l. <http://www.zeroincombenze.org>)
#
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).
#
#    All Rights Reserved
#
"""Clodoo common library
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from future import standard_library

# from builtins import chr
# from builtins import str
# from builtins import *                                           # noqa: F403
from past.builtins import unicode, basestring
import argparse

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
import inspect
import os
import re
import time
import subprocess
from datetime import date

from os0 import os0

standard_library.install_aliases()  # noqa: E402

# Apply for configuration file (True/False)
APPLY_CONF = True
# Default configuration file (i.e. myfile.conf or False for default)
CONF_FN = "./clodoo.conf"
# Read Odoo configuration file (False or /etc/odoo-server.conf)
ODOO_CONF = [
    "/etc/odoo/odoo-server.conf",
    "/etc/odoo/odoo.conf",
    "/etc/odoo-server.conf",
    "/etc/odoo.conf",
    "/etc/openerp/openerp-server.conf",
    "/etc/openerp-server.conf",
    "/etc/odoo/openerp-server.conf",
]
# Read Odoo configuration file (False or /etc/openerp-server.conf)
OE_CONF = False
# Warning: if following LX have no values LX=(), if have 1 value LX=(value,)
# list of string parameters in [options] of config file
LX_CFG_S = (
    "db_name",
    "db_user",
    "db_password",
    "login_user",
    "login2_user",
    "crypt_password",
    "login_password",
    "crypt2_password",
    "login2_password",
    "admin_passwd",
    "db_host",
    "db_port",
    "data_dir",
    "http_port",
    "xmlrpc_port",
    "oe_version",
    "zeroadm_mail",
    "zeroadm_login",
    "oneadm_mail",
    "oneadm_login",
    "oneadm_pwd",
    "botadm_mail",
    "botadm_login",
    "botadm_pwd",
    "svc_protocol",
    "dbfilter",
    "dbfilterd",
    "dbfiltert",
    "dbfilterz",
    "dbtypefilter",
    "companyfilter",
    "userfilter",
    "lang",
    "adm_uids",
    "data_path",
    "date_start",
    "date_stop",
    "account_code",
    "actions",
    "actions_db",
    "actions_mc",
    "actions_uu",
    "heavy_trx",
    "install_modules",
    "uninstall_modules",
    "purge_modules",
    "upgrade_modules",
    "data_selection",
    "modules_2_manage",
    "chart_of_account",
    "catalog_db",
    "custom_act",
    "model",
    "model_code",
    "model_name",
    "model_action",
    "model_keyids",
    "alias_model2",
    "alias_field",
    "hide_cid",
    "modelA",
    "modelA_code",
    "modelA_name",
    "modelA_action",
    "modelA_keyids",
    "aliasA_model2",
    "aliasA_field",
    "hideA_cid",
    "modelB",
    "modelB_code",
    "modelB_name",
    "modelB_action",
    "modelB_keyids",
    "aliasB_model2",
    "aliasB_field",
    "hideB_cid",
    "filename",
    "psycopg2",
    "TRANSDICT",
)
# list of string/boolean parameters in [options] of config file
# Must be declared in LX_CFG_S
LX_CFG_SB = (
    "install_modules",
    "uninstall_modules",
    "purge_modules",
    "actions",
    "actions_db",
    "actions_mc",
    "actions_uu",
    "heavy_trx",
    "psycopg2",
)
# list of pure boolean parameters in [options] of config file
LX_CFG_B = (
    "set_passepartout",
    "check_balance",
    "with_demo",
    "no_fvalidation",
    "draft_recs",
    "setup_banks",
    "setup_account_journal",
    "setup_partners",
    "setup_partner_banks",
    "check_config",
    "exit_onerror",
)
# list of string parameters in both [options] of config file and line command
# or else are just in line command
LX_OPT_S = (
    "db_name",
    "dbfilter",
    "dbg_mode",
    "do_sel_action",
    "dry_run",
    "lang",
    "with_demo",
    "no_fvalidation",
    "lgi_user",
    "lgi_pwd",
    "logfn",
    "quiet_mode",
    # "http_port",
    # "xmlrpc_port",
    "odoo_vid",
    "exit_onerror",
    "data_selection",
)
# List of pure boolean parameters in line command; may be in LX_CFG_S list too
LX_OPT_B = ("dry_run", "with_demo", "no_fvalidation", "exit_onerror")
# List of numeric parameters in line command; may be in LX_CFG_S list too
LX_OPT_N = ()
# list of opponent options
LX_OPT_OPPONENT = {}
# List of string/boolean parameters; may be string or boolean value;
# must be declared in LX_CFG_S or LX_OPT_S
LX_SB = ()
# switch values of options
LX_OPT_ARGS = {}
DEFDCT = {}

msg_time = time.time()


__version__ = "2.0.7"


#############################################################################
# Message and output
#
def init_logger(ctx):
    if ctx["quiet_mode"]:
        os0.set_tlog_file(ctx["logfn"], echo=False)
    else:
        os0.set_tlog_file(ctx["logfn"], echo=True)


def msg_burst(level, text, i, n):
    """Show a message per second from burst sequence as a gauge"""
    global msg_time
    t = time.time() - msg_time
    if t > 1:
        ident = " " * level
        print("\x1b[A{0}[{1:>6}/{2:>6}] {3}".format(ident, i, n, tounicode(text)))
        msg_time = time.time()


def msg(level, text):
    """Show and log a message"""
    ident = " " * level
    txt = "{0}{1}".format(ident, tounicode(text))
    print(txt)


def msg_log(ctx, level, text):
    """Log a message and show if needed"""
    ident = " " * level
    if ctx:
        if "test_unit_mode" in ctx:
            return
        elif ctx["dry_run"] and level > 0:
            txt = "{0}({1})".format(ident, tounicode(text))
        else:
            txt = "{0}{1}".format(ident, tounicode(text))
    else:
        txt = "{0}{1}".format(ident, tounicode(text))
    os0.wlog(txt)


def debug_msg_log(ctx, level, text):
    """Log a debug message and show if needed"""
    global db_msg_sp, db_msg_stack
    # if level == -999:
    #     db_msg_sp += 1
    #     return
    ident = " " * abs(level)
    if ctx.get("dbg_mode", False):
        if "test_unit_mode" in ctx:
            return
        elif ctx["dry_run"] and level > 0:
            txt = ">%s(%s)" % (ident, tounicode(text))
        else:
            txt = ">%s%s" % (ident, tounicode(text))
        print(txt)
        # os0.wlog(txt)


def crypt(s):
    k = "Caserta1959!TO61TV"
    t = ""
    for i, c in enumerate(s):
        j = ord(k[i])
        x = (((ord(c) - 33) + j - (i * 3)) % 95) + 33
        t = t + chr(x)
    return t


def decrypt(t):
    k = "Caserta1959!TO61TV"
    s = ""
    for i, c in enumerate(t):
        j = ord(k[i])
        x = (((ord(c) - 33) - j + (i * 3)) % 95) + 33
        s = s + chr(x)
    return s


def ismbcs(t):
    """ "Return true id string contains mbcs"""
    if isinstance(t, str):
        try:
            t = unicode(t)
            return False
        except BaseException:
            return True
    return False


def strtype(t):
    """Return string type: ascii, mbcs or unicode"""
    if isinstance(t, unicode):
        return "unicode"
    elif isinstance(t, str):
        try:
            t = unicode(t)
            return "ascii"
        except BaseException:
            return "mbcs"
    else:
        return None


def tounicode(s):
    """Return unicode string from basestring"""
    if strtype(s) == "unicode":
        return s
    elif strtype(s) == "mbcs":
        return s.decode("utf-8")
    elif strtype(s) == "ascii":
        return unicode(s)
    else:
        return s


#############################################################################
def default_conf(ctx):
    """Default configuration values"""
    y = date.today().year
    dfmt = "%Y-%m-%d"
    dts_start = date(y, 1, 1).strftime(dfmt)
    dts_stop = date(y, 12, 31).strftime(dfmt)
    return {
        "login_user": "admin",
        "crypt_password": "Ec{fu",
        "login_password": "",
        "login2_user": "admin",
        "crypt2_password": "",
        "login2_password": "",
        "admin_passwd": "admin",
        "db_user": "postgres",
        "db_host": "localhost",
        "data_dir": "",
        "db_port": 5432,
        "oe_version": "*",
        "svc_protocol": "",
        # "http_port": 8069,
        # "odoo_vid": "12.0",
        "db_name": "demo",
        "logfile": False,
        "dbfilter": ".*",
        "dbfilterd": "demo",
        "dbfiltert": "(openerp|odoo|test)",
        "dbfilterz": "zi[0-9]{8}",
        "dbtypefilter": "",
        "companyfilter": r"(?![Zz]eroincombenze.*)",
        "userfilter": ".*",
        "lang": "en_US",
        "with_demo": "0",
        "date_start": dts_start,
        "date_stop": dts_stop,
        "draft_recs": "0",
        "account_code": "000000",
        "adm_uids": "1",
        "set_passepartout": "0",
        "check_balance": "0",
        "setup_banks": "0",
        "setup_account_journal": "0",
        "setup_partners": "0",
        "setup_partner_banks": "0",
        "check_config": "0",
        "exit_onerror": "0",
        "custom_act": "",
        "install_modules": False,
        "uninstall_modules": False,
        "purge_modules": False,
        "upgrade_modules": False,
        "data_selection": "account_move,sale,purchase,project,mail,crm,"
        "inventory,marketing,hr,analytic",
        "modules_2_manage": "",
        "zeroadm_mail": "cc@shs-av.com",
        "zeroadm_login": "zeroadm",
        "oneadm_mail": "admin@example.com",
        "oneadm_login": "admin",
        "oneadm_pwd": "admin",
        "botadm_mail": "zerobot@example.com",
        "botadm_login": "zerobot",
        "botadm_pwd": "",
        "data_path": "./data",
        "actions": "",
        "actions_db": "",
        "actions_mc": "",
        "actions_uu": "",
        "heavy_trx": False,
        "chart_of_account": "configurable_chart_template",
        "catalog_db": "zeroincombenze",
        "psycopg2": "False",
        "caller": "",
        "level": 4,
        "dry_run": False,
        "multi_user": False,
        "ena_inquire": False,
        "no_login": False,
        "TRANSDICT": {},
    }


def get_versioned_option(conf_obj, sect, param, is_bool=None, defval=None):
    is_bool = is_bool or False
    found = False
    if conf_obj:
        for sfx in (
            "6.1",
            "7.0",
            "8.0",
            "9.0",
            "10.0",
            "11.0",
            "12.0",
            "13.0",
            "14.0",
            "15.0",
            "16.0",
        ):
            vparam = "%s_%s" % (param, sfx)
            if conf_obj.has_option(sect, vparam):
                found = True
                break
    if not found:
        vparam = param
        if conf_obj and conf_obj.has_option(sect, vparam):
            found = True
    if found:
        if is_bool:
            return conf_obj.getboolean(sect, vparam)
        value = conf_obj.get(sect, vparam)
        if value != "False":
            return value
    if defval and param in defval:
        return defval[param]
    else:
        return None


def create_def_params_dict(ctx):
    """Create default params dictionary"""
    opt_obj = ctx.get("_opt_obj", None)
    conf_obj = ctx.get("_conf_obj", None)
    s = "options"
    if conf_obj and not conf_obj.has_section(s):
        conf_obj.add_section(s)
    DEFDCT = default_conf(ctx)
    for p in LX_CFG_S:
        v = get_versioned_option(conf_obj, s, p, defval=DEFDCT)
        if v is not None:
            ctx[p] = v
    for p in LX_CFG_B:
        v = get_versioned_option(conf_obj, s, p, is_bool=True)
        if v is not None:
            ctx[p] = v
    if opt_obj:
        for p in LX_OPT_S:
            if p in LX_OPT_OPPONENT:
                a = LX_OPT_OPPONENT[p]
                if hasattr(opt_obj, a) and getattr(opt_obj, a) is False:
                    ctx[p] = False
                elif hasattr(opt_obj, p) and getattr(opt_obj, p):
                    ctx[p] = True
                else:
                    ctx[p] = None
            elif hasattr(opt_obj, p):
                tmp = getattr(opt_obj, p)
                if p not in ctx or tmp:
                    ctx[p] = tmp
        for p in LX_OPT_B:
            if hasattr(opt_obj, p):
                ctx[p] = os0.str2bool(getattr(opt_obj, p), None)
        for p in LX_OPT_N:
            if hasattr(opt_obj, p) and getattr(opt_obj, p):
                ctx[p] = int(getattr(opt_obj, p))
    for p in LX_CFG_SB:
        ctx[p] = os0.str2bool(ctx[p], ctx[p])
    if ctx.get("LX_CFG_S", ""):
        ctx["LX_CFG_S"] = eval(ctx["LX_CFG_S"])
    return ctx


def create_params_dict(ctx):
    """Create all params dictionary"""
    ctx = create_def_params_dict(ctx)
    DEFDCT = default_conf(ctx)
    if ctx.get("dbg_mode", None) is None:
        ctx["dbg_mode"] = ctx["run_daemon"]
    if not ctx.get("logfn", None):
        if "tlog" in ctx:
            ctx["logfn"] = ctx["tlog"]
        else:
            ctx["logfn"] = "~/" + ctx["caller"] + ".log"
    conf_obj = ctx.get("_conf_obj", None)
    opt_obj = ctx.get("_opt_obj", None)
    s = "options"
    if conf_obj and not conf_obj.has_section(s):
        conf_obj.add_section(s)
    v = get_versioned_option(conf_obj, s, "db_password", defval=DEFDCT)
    if v is not None:
        ctx["db_pwd"] = v
    else:
        ctx["db_pwd"] = ""
    for p in ():
        ctx[p] = conf_obj.getint(s, p)
    if opt_obj:
        if hasattr(opt_obj, "dbfilter") and opt_obj.dbfilter != "":
            ctx["dbfilter"] = opt_obj.dbfilter
            ctx["multi_db"] = True
        if hasattr(opt_obj, "modules_2_manage") and hasattr(opt_obj, "do_sel_action"):
            if opt_obj.do_sel_action in (
                "install_modules",
                "uninstall_modules",
                "upgrade_modules",
                "purge_modules",
            ):
                ctx[opt_obj.do_sel_action] = opt_obj.modules_2_manage
        if hasattr(opt_obj, "data_path") and opt_obj.data_path != "":
            ctx["data_path"] = opt_obj.data_path
    if ctx.get("oe_version") and ctx["oe_version"] != "*" and not ctx.get("odoo_vid"):
        ctx["odoo_vid"] = ctx["oe_version"]
    elif ctx.get("odoo_vid") and ctx["odoo_vid"] != "*":
        ctx["oe_version"] = build_odoo_param("FULLVER", ctx["odoo_vid"])
    if not ctx["svc_protocol"]:
        if ctx["oe_version"] in ("9.0", "8.0", "7.0", "6.1"):
            ctx["svc_protocol"] = "xmlrpc"
        else:
            ctx["svc_protocol"] = "jsonrpc"
    if ctx.get("do_sel_action", False):
        ctx["actions"] = ctx["do_sel_action"]
    elif ctx.get("actions_db", None):
        ctx["actions"] = "per_db," + ctx["actions_db"]
        del ctx["actions_db"]
    elif ctx.get("actions_mc", None):
        ctx["actions"] = "per_company," + ctx["actions_mc"]
        del ctx["actions_mc"]
    elif ctx.get("actions_uu", None):
        ctx["actions"] = "per_users," + ctx["actions_uu"]
        del ctx["actions_uu"]
    for p in ("http_port", "xmlrpc_port", "db_port"):
        if isinstance(ctx.get(p), basestring) and ctx[p].isdigit():
            ctx[p] = int(ctx[p])
    return ctx


def docstring_summary(docstring):
    """Return summary of docstring."""
    for text in docstring.split("\n"):
        if text.strip():
            break
    return text.strip()


def fullname_conf(ctx):
    if ctx.get("conf_fn"):
        confs = []
        for confn in ctx["conf_fn"].split(","):
            if not os.path.isfile(confn):
                for path in (".", "./confs", "./conf", "./code"):
                    if os.path.isfile("%s/%s" % (path, confn)):
                        confn = "%s/%s" % (path, confn)
                        break
            confs.append(confn)
        if len(confs) == 1:
            ctx["conf_fn"] = confs[0]
        else:
            ctx["conf_fn"] = ",".join(confs)
    return ctx


def read_config(ctx):
    """Read both user configuration and local configuration."""

    def set_confs(ctx):
        version_is_set = False
        for ver in ("odoo_vid", "oe_version"):
            if ctx.get(ver) and ctx[ver] != "*":
                fnver = build_odoo_param("CONFN", ctx[ver], multi=True)
                if os.path.isfile(fnver) and fnver not in ctx["conf_fns"]:
                    ctx["conf_fns"].insert(0, fnver)
                    version_is_set = True
                else:
                    fnver = build_odoo_param("CONFN", ctx[ver])
                    if os.path.isfile(fnver) and fnver not in ctx["conf_fns"]:
                        ctx["conf_fns"].insert(0, fnver)
                        version_is_set = True
        return ctx, version_is_set

    if not ctx.get("conf_fn", None):
        ctx["conf_fn"] = ctx.get("caller", "clodoo") + ".conf"
    ctx = fullname_conf(ctx)
    ctx["conf_fns"] = ctx["conf_fn"].split(",")
    ctx, version_is_set = set_confs(ctx)
    ctx["_conf_obj"] = ConfigParser.RawConfigParser(default_conf(ctx))
    ctx["conf_fns"] = ctx["_conf_obj"].read(ctx["conf_fns"])
    if not ctx["conf_fns"]:
        raise IOError("No configuration file found!")
    ctx = create_params_dict(ctx)
    if not version_is_set:
        ctx, version_is_set = set_confs(ctx)
        if version_is_set:
            ctx["conf_fns"] = ctx["_conf_obj"].read(ctx["conf_fns"])
            ctx = create_params_dict(ctx)
    if (not ctx.get("http_port") and not ctx.get("xmlrpc_port")):
        if int(ctx.get("oe_version", "12.0").split(".")[0]) < 10:
            ctx["xmlrpc_port"] = 8069
        else:
            ctx["http_port"] = 8069
    return ctx


def create_parser(version, doc, ctx):
    """Return command-line parser.
    Some options are standard:
    -c --config     set configuration file (conf_fn)
    -h --help       show help
    -q --quiet      quiet mode
    -t --dry-run    simulation mode for test (simulate)
    -U --user       set username (user)
    -v --verbose    verbose mode (dbg_mode)
    -V --version    show version
    -y --yes        confirmation w/out ask
    """
    parser = argparse.ArgumentParser(
        description=docstring_summary(doc),
        epilog="© 2015-2021 by SHS-AV s.r.l." " - http://www.zeroincombenze.org",
    )
    parser.add_argument(
        "-A",
        "--action-to-do",
        help="action to do (use list_actions to dir)",
        dest="do_sel_action",
        metavar="actions",
        default=None,
    )
    parser.add_argument(
        "-b",
        "--odoo-branch",
        help="talk server Odoo version",
        dest="odoo_vid",
        metavar="version",
        default="",
    )
    parser.add_argument(
        "-c",
        "--config",
        help="configuration command file",
        dest="conf_fn",
        metavar="file",
        default=CONF_FN,
    )
    parser.add_argument(
        "-d",
        "--dbfilter",
        help="DB filter",
        dest="dbfilter",
        metavar="regex",
        default="",
    )
    parser.add_argument(
        "-i",
        "--ignore-name-validation",
        help="ignore name validation fo csv columns",
        action="store_true",
        dest="no_fvalidation",
        default=False,
    )
    parser.add_argument(
        "-l",
        "--lang",
        help="user language",
        dest="lang",
        metavar="iso_lang",
        default=False,
    )
    parser.add_argument(
        "-m",
        "--modules-2-manage",
        help="Module list to upgrade",
        dest="modules_2_manage",
        metavar="list",
        default="",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        help="test execution mode",
        action="store_true",
        dest="dry_run",
        default=False,
    )
    parser.add_argument(
        "-o",
        "--with-demo",
        help="create db with demo data",
        action="store_true",
        dest="with_demo",
        default=False,
    )
    parser.add_argument(
        "-p",
        "--data-path",
        help="Import file path",
        dest="data_path",
        metavar="dir",
        default="",
    )
    parser.add_argument(
        "-P",
        "--pwd",
        help="login password",
        dest="lgi_pwd",
        metavar="password",
        default="admin",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        help="run silently",
        action="store_true",
        dest="quiet_mode",
        default=False,
    )
    parser.add_argument(
        "-r",
        "--http-port",
        help="http / xmlrpc port",
        dest="http_port",
        metavar="port",
        default="",
    )
    parser.add_argument(
        "-S",
        "--data-selection",
        help="Select data to remove",
        dest="data_selection",
        metavar="list",
        default="",
    )
    parser.add_argument(
        "-U",
        "--user",
        help="login username",
        dest="lgi_user",
        metavar="username",
        default=None,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="run with debugging output",
        action="store_true",
        dest="dbg_mode",
        default=False,
    )
    parser.add_argument(
        "-V", "--version", action="version", version="%(prog)s " + version
    )
    parser.add_argument(
        "-x",
        "--exit-on-error",
        help="exit on error",
        action="store_true",
        dest="exit_onerror",
        default=False,
    )
    return parser


def set_base_ctx(ctx=None):
    ctx = ctx or {}
    ctx["caller_fqn"] = inspect.stack()[1][1]
    ctx["caller"] = os0.nakedname(os.path.basename(ctx["caller_fqn"]))
    ctx["run_daemon"] = False if os.isatty(0) else True
    ctx["run_tty"] = os.isatty(0)
    return ctx


def parse_args(arguments, apply_conf=APPLY_CONF, version=None, tlog=None, doc=None):
    """Parse command-line options."""
    ctx = set_base_ctx()
    if tlog:
        ctx["tlog"] = tlog
    else:
        ctx["tlog"] = "./%s.log" % ctx["caller"]
    # running autotest
    if version is None:
        ctx["_run_autotest"] = True
    parser = create_parser(version, doc, ctx)
    ctx["_parser"] = parser
    opt_obj = parser.parse_args(arguments)
    ctx["_opt_obj"] = opt_obj
    if apply_conf:
        if hasattr(opt_obj, "conf_fn"):
            ctx["conf_fn"] = opt_obj.conf_fn
        if hasattr(opt_obj, "odoo_vid") and opt_obj.odoo_vid:
            ctx["odoo_vid"] = opt_obj.odoo_vid
        ctx = read_config(ctx)
        opt_obj = parser.parse_args(arguments)
    ctx["level"] = 0
    ctx = create_params_dict(ctx)
    return ctx


def check_if_running(ctx, pid):
    f_alrdy_run = False
    id_str = ctx["caller"] + ".py"
    cmd = "ps aux|grep " + id_str
    os0.muteshell(cmd, keepout=True)
    stdinp_fd = open(os0.setlfilename(os0.bgout_fn), "r")
    rxmatch = "root .* python .*" + id_str + ".*"
    rxnmatch = "root .* {0} .*".format(pid)
    line = stdinp_fd.readline()
    while line != "" and not f_alrdy_run:
        i = line.rfind("\n")
        if i >= 0:
            if re.match(rxmatch, line) and not re.match(rxnmatch, line):
                f_alrdy_run = True
        line = stdinp_fd.readline()
    stdinp_fd.close()
    if os.path.isfile(os0.setlfilename(os0.bgout_fn)):
        os.remove(os0.setlfilename(os0.bgout_fn))
    return f_alrdy_run


def get_odoo_full_ver(odoo_vid):
    v = re.search(r"[0-9]+(\.[0-9])?", odoo_vid).group()
    if v == "6":
        odoo_fver = "6.1"
    elif v.find(".") >= 0:
        odoo_fver = v
    else:
        odoo_fver = v + ".0"
    return odoo_fver


def build_odoo_VER(
    item, odoo_vid=None, debug=None, suppl=None, git_org=None, multi=None
):
    return __version__


def build_odoo_param(
    item, odoo_vid=None, debug=None, suppl=None, git_org=None, multi=None
):
    fct = "build_odoo_%s" % item
    if fct in globals():
        return globals()[fct](
            odoo_vid=odoo_vid, debug=debug, suppl=suppl, git_org=git_org, multi=multi
        )
    odoorc = os.path.join(os.path.dirname(__file__), "odoorc")
    if multi:
        cmd = 'opt_multi=1 %s %s "%s" "%s" "%s"' % (
            odoorc,
            item,
            odoo_vid,
            suppl or "",
            git_org or "",
        )
    else:
        cmd = '%s %s "%s" "%s" "%s"' % (
            odoorc,
            item,
            odoo_vid,
            suppl or "",
            git_org or "",
        )
    out, err = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()
    if not out:
        return False
    out = os0.u(out)
    if item in ("LPPORT", "MAJVER", "RPCPORT"):
        return eval(out.split("\n")[0])
    return out.split("\n")[0]
